/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Lukasz Kiedrowski <kiedrowski@itti.com.pl>
 * Copyright (c) ITTI sp. z o.o. 2011
 */
#include <ctime>
//n4c library
#include "ns3/ConfigFile.h"
#include "ns3/dtn-tool.h"
#include "ns3/visualizer.h" //by default not present in 3.10; moved to n4c/visualizer (based on one from ver. 3.9)
//other
#include "ns3/simulator-module.h"
#include "ns3/node-module.h"
#include "ns3/core-module.h"
#include "ns3/wifi-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-model.h"
#include "ns3/mobility-module.h"
#include "ns3/topology-read-module.h"

using namespace ns3;
using namespace std;

string sim_name = "GraphRnd_v2";

NS_LOG_COMPONENT_DEFINE (sim_name.c_str ());


void Change_node_attr (DTNTool* anim_helper)
{
	anim_helper->ChangeNodeAttr ();
}
void Change_node_pos_vel (const NodeContainer* nc, DTNTool* anim_helper)
{
	anim_helper->CourseChangedVelocity (nc);
}

int 
main (int argc, char *argv[])
{
  LogComponentEnable(sim_name.c_str (), LOG_LEVEL_INFO);
//VARIABLES
  CommandLine cmd;
  ostringstream ossn, ossmsg;
  double rt = 600;
  double uti = 0.5;
  double circle_radius = 40;
  uint32_t vbox_no = 2;
  uint32_t lxc_no = 3;
  uint32_t lxc_mob_no = 1;
  uint32_t vpn_no = 1;
  bool pyviz = false;
//---TabBridge
  string tbm = "UseLocal";
  string iprefix = "i";
//---Animation
  uint16_t    port = 0;  // If non zero, port no. to bind to connection to NetAnim
  string animFile = "../n4c_config/NetAnim/traces/";  // Name of file for 
  DTNTool anim_helper;
  DTNTool* anim_helper_ptr = &anim_helper;
//---Interactive console
  string path_interactive = "../n4c_config/interactive/";
  string user = "lukasz";
//---Topology
  //string format ("ns2");
  string input ("../n4c_config/ns3_topology_generation/adhoc_tg/topologies/test5");

//CONFIGURATION
//---Read configuration file
  string conf = "test.conf";
  cmd.AddValue("conf", "Path and filename of config file", conf);
  cmd.Parse (argc, argv);
  ifstream conffile(conf.c_str ());
if (conffile){
  ConfigFile config(conf);

  lxc_no = config.read<uint32_t>( "lxc_no", 1 );
  lxc_mob_no = config.read<uint32_t>( "lxc_mob_no", 1 );
  vbox_no = config.read<uint32_t>( "vbox_no", 1 );
  vpn_no = config.read<uint32_t>( "vpn_no", 1 );
  rt = config.read<double>( "rt", 600.0 );
  uti = config.read<double>( "uti", 0.5 );
  circle_radius = config.read<double>( "circle_radius" );
  port = config.read<uint32_t>( "port", 0 ); 
  animFile = config.read<string>( "anim_path", "../n4c_config/NetAnim/traces/" );
  input = config.read<string>( "input_top", "../n4c_config/ns3_topology_generation/adhoc_tg/topologies/test5" ); 
  path_interactive = config.read<string>( "path_interactive", "../n4c_config/interactive/" );
  user = config.read<string>( "user", "lukasz" );
  pyviz = config.read<bool>( "pyviz", false ); 
}
else
{
  NS_LOG_INFO("No file or directory (configuration file):" << conf.c_str ());
  return 1;
}
  NS_LOG_INFO("Number of vbox nodes - " << vbox_no);
  NS_LOG_INFO("Number of vpn nodes - " << vpn_no);
  NS_LOG_INFO("Number of lxc nodes - " << lxc_no);
  NS_LOG_INFO("Port Number for NetAnim " << port);

//INITIALIZATION - n4c tools
//---Read topology spacepoints
  Vector3D* sp_tpg=NULL;

  ifstream mobfile(input.c_str ());
  if (mobfile) {
    // The file exists, and is open for input
    sp_tpg = anim_helper.ReadPositions (input);
  }else
  {
    NS_LOG_INFO("No file or directory " << input.c_str ());
    return 1;
  }

  Ptr<ListPositionAllocator> pa_sta = CreateObject<ListPositionAllocator> ();
  int z;
  for (z=0;z<100 && z<int(lxc_no-lxc_mob_no);z++)
  {
   //NS_LOG_INFO("      " << sp_tpg[z]);
   pa_sta->Add (sp_tpg[z]);
  }
//----Movement logging or remote NetAnim animation
  if (port > 0) anim_helper.SetServerPort (port);
  else 
  {
     ossn.str("");
     ossn << animFile.c_str () << sim_name.c_str ();
     animFile = ossn.str ();
     anim_helper.SetOutputFile (animFile);
  }
//----Initialization and starting interactive terminal
  anim_helper.InitializeInteractiveTerminal (user, path_interactive);
//----Setting transmision range 
  anim_helper.SetRadius (circle_radius);

//INITIALIZATION - NS3
  GlobalValue::Bind ("SimulatorImplementationType", StringValue ("ns3::RealtimeSimulatorImpl"));
  GlobalValue::Bind ("ChecksumEnabled", BooleanValue (true));

  NodeContainer nodes_lxc_sta;
  NodeContainer nodes_lxc_mob;
  NodeContainer* nc_lxc_mob = &nodes_lxc_mob;
  NodeContainer nodes_vpn;
  NodeContainer nodes_vbox;

  //uint32_t nodes_total = lxc_no+vpn_no+vbox_no;
  //DO NOT change the order of nodes instalation (creation)!!!
  if (vbox_no >0) nodes_vbox.Create (vbox_no);
  if (vpn_no >0) nodes_vpn.Create (vpn_no);
  if (lxc_no >0) nodes_lxc_sta.Create (lxc_no-lxc_mob_no);
  if (lxc_mob_no >0) nodes_lxc_mob.Create (lxc_mob_no);

  NS_LOG_INFO("Number of created nodes: " << lxc_no+vpn_no+vbox_no);
//---

  WifiHelper wifi = WifiHelper::Default ();
  //wifi.SetStandard (WIFI_PHY_STANDARD_80211a);
  //wifi.SetRemoteStationManager ("ns3::ConstantRateWifiManager", "DataMode", StringValue ("wifia-54mbs"));

  YansWifiChannelHelper wifiChannel = YansWifiChannelHelper::Default (); 
  YansWifiPhyHelper wifiPhy = YansWifiPhyHelper::Default ();
  NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
  wifiMac.SetType ("ns3::AdhocWifiMac");
  //
  ObjectFactory factory;
  factory.SetTypeId ("ns3::RangePropagationLossModel");
  factory.Set ("MaxRange", DoubleValue (circle_radius));
  //Ptr<PropagationLossModel> cur = (*i).Create<PropagationLossModel> ();
  Ptr<PropagationLossModel> simple = factory.Create<PropagationLossModel> ();

  //wifiPhy.SetChannel (wifiChannel.Create ());
  Ptr<YansWifiChannel> channel = wifiChannel.Create ();
  channel->SetPropagationLossModel (simple);
  wifiPhy.SetChannel (channel);

  NetDeviceContainer dev_vbox;
  NetDeviceContainer dev_vpn;
  NetDeviceContainer dev_lxc_sta;
  NetDeviceContainer dev_lxc_mob;

  if (vbox_no >0) dev_vbox = wifi.Install (wifiPhy, wifiMac, nodes_vbox);
  if (vpn_no >0) dev_vpn = wifi.Install (wifiPhy, wifiMac, nodes_vpn);
  if (lxc_no >0) dev_lxc_sta = wifi.Install (wifiPhy, wifiMac, nodes_lxc_sta);
  if (lxc_mob_no >0) dev_lxc_mob = wifi.Install (wifiPhy, wifiMac, nodes_lxc_mob);

  MobilityHelper mobility;
  Ptr<ListPositionAllocator> pa_init = CreateObject<ListPositionAllocator> ();
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");

  TapBridgeHelper tapBridge;
  tapBridge.SetAttribute("Mode", StringValue(tbm));
  // 
  NS_LOG_INFO("Spacepoints created");

  uint32_t k;
  for (k=0;k<vbox_no;k++)
  {
   pa_init->Add (sp_tpg[(lxc_no-lxc_mob_no)+k]); 
   NS_LOG_INFO("      " <<"vbox vm "<< iprefix << k+1 << " at " << sp_tpg[(lxc_no-lxc_mob_no)+k]);
   ossn.str("");
   ossn << iprefix << k+1;
   tapBridge.SetAttribute("DeviceName", StringValue(ossn.str()));
   tapBridge.Install(nodes_vbox.Get(k), dev_vbox.Get(k));
  }
  for (k=0;k<vpn_no;k++) 
  {
   pa_init->Add (sp_tpg[(lxc_no-lxc_mob_no)+vbox_no+k]); 
   NS_LOG_INFO("      " <<"vpn vm "<< iprefix << k+201 << " at " << sp_tpg[(lxc_no-lxc_mob_no)+vbox_no+k]);
   ossn.str("");
   ossn << iprefix << k+201;
   tapBridge.SetAttribute("DeviceName", StringValue(ossn.str()));
   tapBridge.Install(nodes_vpn.Get(k), dev_vpn.Get(k));
  }
  for (k=0;k<lxc_no;k++)//positions between first and last
  {
   //positionAlloc->Add (sp_tpg[k]);
   //NS_LOG_INFO("      " <<"lxc vm "<< iprefix << k+101 << " at " << sp_tpg[k]);
   ossn.str("");
   ossn << iprefix << k+101;
   tapBridge.SetAttribute("DeviceName", StringValue(ossn.str()));
   if (k < (lxc_no-lxc_mob_no)){
     tapBridge.Install(nodes_lxc_sta.Get(k), dev_lxc_sta.Get(k));
   }
   else{
     tapBridge.Install(nodes_lxc_mob.Get(k-(lxc_no-lxc_mob_no)), dev_lxc_mob.Get(k-(lxc_no-lxc_mob_no))); 
   }
  }
  //DO NOT change the order of nodes position allocation!!!
  mobility.SetPositionAllocator (pa_init);
  if (vbox_no >0) mobility.Install (nodes_vbox);
  if (vpn_no >0) mobility.Install (nodes_vpn);

  mobility.SetPositionAllocator (pa_sta);
  if (lxc_no >0) mobility.Install (nodes_lxc_sta);

  Ptr<RandomRectanglePositionAllocator> pa_mob = CreateObject<RandomRectanglePositionAllocator> ();
  double minX=350.0; 
  double maxX=625.0;
  double minY=350.0;
  double maxY=600.0;
  pa_mob->SetX (UniformVariable (minX, maxX));
  pa_mob->SetY (UniformVariable (minY, maxY));
  mobility.SetPositionAllocator (pa_mob);

 /* mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
    "Bounds", RectangleValue (Rectangle (350.0, 625.0, 350.0, 650.0)),
    "Distance", DoubleValue(5.0), "Time", StringValue("1000000.0ns"));
  mobility.Install (nodes_lxc_mob);
*/
  mobility.SetMobilityModel ("ns3::RandomDirection2dMobilityModel",
    "Bounds", RectangleValue (Rectangle (350.0, 625.0, 350.0, 600.0)), // 
    "Speed", RandomVariableValue (ConstantVariable (15)),  //RandomVariableValue (UniformVariable (8, 10)), 
    "Pause", RandomVariableValue (ConstantVariable (0.2)));
  if (lxc_mob_no >0) mobility.Install (nodes_lxc_mob);

  wifiPhy.EnablePcapAll(sim_name.c_str (), true); // all packet captured

//STARTING REMOTE ANIMATION
  anim_helper.StartAnimationWifi (vbox_no, vpn_no, lxc_no, 0);

//SCHEDULING EVENTS 
  for (uint32_t ts=1; ts<rt; ts++)
  {
    Simulator::Schedule(Seconds(ts), &Change_node_attr, anim_helper_ptr); //check if any commmand from interactive terminal
    if (lxc_mob_no >0) Simulator::Schedule(Seconds(double(ts)*uti), &Change_node_pos_vel, nc_lxc_mob, anim_helper_ptr);
  }

  Simulator::Stop(Seconds(rt));
  time_t ct;
  time(&ct);
  cerr << "SIMULATION START " << ctime(&ct);

  if (pyviz) Visualizer::Run (); // wrong in 3.10
  else Simulator::Run(); 
  //Simulator::Run(); 

  time(&ct);
  cerr << "SIMULATION END   " << ctime(&ct);
  Simulator::Destroy();
  cout << "Stopping the animation" << endl;
  anim_helper.StopAnimation();
  return 0;
}
