/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Lukasz Kiedrowski <kiedrowski@itti.com.pl>
 * Copyright (c) ITTI sp. z o.o. 2010
 */
#include <ctime>
//n4c library
#include "ns3/ConfigFile.h"
#include "ns3/dtn-tool.h"
#include "ns3/visualizer.h" //by default not present in 3.10; moved to n4c/visualizer (based on one from ver. 3.9)
//other
#include "ns3/simulator-module.h"
#include "ns3/node-module.h"
#include "ns3/core-module.h"
#include "ns3/wifi-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-model.h"
#include "ns3/mobility-module.h"
#include "ns3/topology-read-module.h"

using namespace ns3;
using namespace std;

string sim_name = "Inline";

NS_LOG_COMPONENT_DEFINE (sim_name.c_str ());

const uint32_t spmax_lxc = 100;
const uint32_t spmax_vbox = 2;
const uint32_t spmax_vpn = 1;
Vector3D sp_lxc[spmax_lxc];
Vector3D sp_vbox[spmax_vbox];
Vector3D sp_vpn[spmax_vpn];
map<uint32_t, uint32_t> id_to_real; //have made it global 1901
string iprefix = "i";

void move_nodes (const NodeContainer* nc, const int32_t pref, double dist,
                 const bool logpositions)
{
  Vector3D v, y;
  uint32_t i;
  if (logpositions)
    NS_LOG_INFO("  MOVEMENT odd (lxc) nodes");
  for (i=0; i<nc->GetN(); i+=2) //only odd nodes
  {
    v = sp_lxc[i];
    v.x = v.x + pref*dist/2;
    v.y = v.y + pref*dist/2;
    ((nc->Get(i))->GetObject<MobilityModel>())->SetPosition(v);
    y=((nc->Get(i))->GetObject<MobilityModel>())->GetPosition();
    
    if (logpositions)
      NS_LOG_INFO("    " << iprefix << i+101 << "( vm" << id_to_real[(nc->Get(i))->GetId()]<< " )" << " at " << y);
  }
}

void Change_node_attr (DTNTool* anim_helper)
{
	anim_helper->ChangeNodeAttr ();
}

int 
main (int argc, char *argv[])
{
  LogComponentEnable(sim_name.c_str (), LOG_LEVEL_INFO);
//VARIABLES
  CommandLine cmd;
  ostringstream ossn, ossmsg;
  double rt = 600;
  double ti = 5;
  double dist = 70;
  double circle_radius = 40;
  uint32_t vbox_no = 2;
  uint32_t lxc_no = 3;
  uint32_t vpn_no = 1;
  bool pyviz = false;
//---TabBridge
  string tbm = "UseLocal";
//---Animation
  uint16_t    port = 0;  // If non zero, port to bind to for anim connection
  string animFile = "../n4c_config/NetAnim/traces/";  // Name of file for 
  DTNTool anim_helper;
  DTNTool* anim_helper_ptr = &anim_helper;
//---Interactive console
  string path_interactive = "../n4c_config/interactive/";
  string user = "lukasz";

//CONFIGURATION
//---Read configuration file
  string conf = "test.conf";
  cmd.AddValue("conf", "Path and filename of config file", conf);
  cmd.Parse (argc, argv);
  ifstream conffile(conf.c_str ());
if (conffile){
  ConfigFile config(conf);

  lxc_no = config.read<uint32_t>( "lxc_no", 3 ); //must be odd
  //vbox_no = config.read<uint32_t>( "vbox_no", 2 );  // popraw
  //vpn_no = config.read<uint32_t>( "vpn_no", 0 );	// popraw
  rt = config.read<double>( "rt", 600.0 );
  ti = config.read<double>( "ti", 5.0 );
  dist = config.read<double>( "dist", 50.0 );
  circle_radius = config.read<double>( "circle_radius" );
  port = config.read<uint32_t>( "port", 0 ); 
  animFile = config.read<string>( "anim_path", "../n4c_config/NetAnim/traces/" );
  path_interactive = config.read<string>( "path_interactive", "../n4c_config/interactive/" );
  user = config.read<string>( "user", "lukasz" );
  pyviz = config.read<bool>( "pyviz", false ); 
}
else
{
  NS_LOG_INFO("No file or directory (configuration file):" << conf.c_str ());
  return 1;
}
if (lxc_no%2==0) 
{
 NS_LOG_INFO("Linux containers number lxc_no must be odd");
 return 1;
}
  NS_LOG_INFO("Number of vbox nodes - " << vbox_no);
  NS_LOG_INFO("Number of vpn nodes - " << vpn_no);
  NS_LOG_INFO("Number of lxc nodes - " << lxc_no);
  NS_LOG_INFO("Port number for NetAnim " << port);

//INITIALIZATION - n4c tools
//----Movement logging or remote NetAnim animation
  if (port > 0) anim_helper.SetServerPort (port);
  else 
  {
     ossn.str("");
     ossn << animFile.c_str () << sim_name.c_str ();
     animFile = ossn.str ();
     anim_helper.SetOutputFile (animFile);
  }
//----Initialization and starting interactive terminal
  anim_helper.InitializeInteractiveTerminal (user, path_interactive);
//----Setting transmision range 
  anim_helper.SetRadius (circle_radius);

//INITIALIZATION - NS3
  GlobalValue::Bind ("SimulatorImplementationType", StringValue ("ns3::RealtimeSimulatorImpl"));
  GlobalValue::Bind ("ChecksumEnabled", BooleanValue (true));

  NodeContainer nodes_lxc;
  NodeContainer* nc_lxc = &nodes_lxc;
  NodeContainer nodes_vpn;
  NodeContainer nodes_vbox;
  uint32_t nodes_total = lxc_no+vpn_no+vbox_no;
  //DO NOT change the order of nodes instalation (creation)!!!
  if (vbox_no >0) nodes_vbox.Create (vbox_no);
  if (vpn_no >0) nodes_vpn.Create (vpn_no);
  if (lxc_no >0) nodes_lxc.Create (lxc_no);

  NS_LOG_INFO("Number of created nodes: " << nodes_total);
//---

  WifiHelper wifi = WifiHelper::Default ();
  //wifi.SetStandard (WIFI_PHY_STANDARD_80211a);
  //wifi.SetRemoteStationManager ("ns3::ConstantRateWifiManager", "DataMode", StringValue ("wifia-54mbs"));

  YansWifiChannelHelper wifiChannel = YansWifiChannelHelper::Default (); 
  YansWifiPhyHelper wifiPhy = YansWifiPhyHelper::Default ();
  NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
  wifiMac.SetType ("ns3::AdhocWifiMac");
  //
  ObjectFactory factory;
  factory.SetTypeId ("ns3::RangePropagationLossModel");
  factory.Set ("MaxRange", DoubleValue (circle_radius));
  //Ptr<PropagationLossModel> cur = (*i).Create<PropagationLossModel> ();
  Ptr<PropagationLossModel> simple = factory.Create<PropagationLossModel> ();

  //wifiPhy.SetChannel (wifiChannel.Create ());
  Ptr<YansWifiChannel> channel = wifiChannel.Create ();
  channel->SetPropagationLossModel (simple);
  wifiPhy.SetChannel (channel);

  NetDeviceContainer dev_vbox;
  NetDeviceContainer dev_vpn;
  NetDeviceContainer dev_lxc;

  if (vbox_no >0) dev_vbox = wifi.Install (wifiPhy, wifiMac, nodes_vbox);
  if (vpn_no >0) dev_vpn = wifi.Install (wifiPhy, wifiMac, nodes_vpn);
  if (lxc_no >0) dev_lxc = wifi.Install (wifiPhy, wifiMac, nodes_lxc);

  MobilityHelper mobility;
  Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");

  TapBridgeHelper tapBridge;
  tapBridge.SetAttribute("Mode", StringValue(tbm));
// 
  uint32_t sp; 
  for (sp=0; sp<vbox_no; sp++){
   sp_vbox[sp] = Vector(sp*(nodes_total-vpn_no-1)*dist, sp*(nodes_total-vpn_no-1)*dist, 0.0);
  }  
  for (sp=0; sp<vpn_no; sp++){
   sp_vpn[sp] = Vector(0.0, sp*nodes_total*dist, 0.0);
  } 
  for (sp=0; sp<lxc_no; sp++){
   sp_lxc[sp] = Vector((sp+1)*dist, (sp+1)*dist, 0.0);
  }
  NS_LOG_INFO("Spacepoints created");

  uint32_t k;
  for (k=0;k<vbox_no;k++)
  {
   positionAlloc->Add (sp_vbox[k]);
   NS_LOG_INFO("      " <<"vbox vm "<< iprefix << k+1 << " at " << sp_vbox[k]);
   ossn.str("");
   ossn << iprefix << k+1;
   tapBridge.SetAttribute("DeviceName", StringValue(ossn.str()));
   tapBridge.Install(nodes_vbox.Get(k), dev_vbox.Get(k));
  }
  for (k=0;k<vpn_no;k++)
  {
   positionAlloc->Add (sp_vpn[k]);
   NS_LOG_INFO("      " <<"vpn vm "<< iprefix << k+201 << " at " << sp_vpn[k]);
   ossn.str("");
   ossn << iprefix << k+201;
   tapBridge.SetAttribute("DeviceName", StringValue(ossn.str()));
   tapBridge.Install(nodes_vpn.Get(k), dev_vpn.Get(k));
  }

  for (k=0;k<lxc_no;k++)
  {
   positionAlloc->Add (sp_lxc[k]);
   NS_LOG_INFO("      " <<"lxc vm "<< iprefix << k+101 << " at " << sp_lxc[k]);
   ossn.str("");
   ossn << iprefix << k+101;
   tapBridge.SetAttribute("DeviceName", StringValue(ossn.str()));
   tapBridge.Install(nodes_lxc.Get(k), dev_lxc.Get(k));
  }
  //DO NOT change the order of nodes position allocation!!!
  mobility.SetPositionAllocator (positionAlloc);
  if (vbox_no >0) mobility.Install (nodes_vbox);
  if (vpn_no >0) mobility.Install (nodes_vpn);
  if (lxc_no >0) mobility.Install (nodes_lxc);

  wifiPhy.EnablePcapAll(sim_name.c_str (), true); // all packet captured

//STARTING REMOTE ANIMATION
  id_to_real = anim_helper.StartAnimationWifi (vbox_no, vpn_no, lxc_no, 0);

//SCHEDULING EVENTS 
  int32_t pref=1;
  for (uint32_t ts=1; ts<(uint32_t(floor(rt/ti))-10); ts++)
  {
    pref=-pref;
    Simulator::Schedule(Seconds(ti*ts), &move_nodes, nc_lxc, pref, dist, true);
    Simulator::Schedule(Seconds(ts), &Change_node_attr, anim_helper_ptr); //check if any commmand from interactive terminal
  }

  Simulator::Stop(Seconds(rt));
  time_t ct;
  time(&ct);
  cerr << "SIMULATION START " << ctime(&ct);

  if (pyviz) Visualizer::Run (); // wrong in 3.10
  else Simulator::Run(); 
  //Simulator::Run(); 

  time(&ct);
  cerr << "SIMULATION END   " << ctime(&ct);
  Simulator::Destroy();
  cout << "Stopping the animation" << endl;
  anim_helper.StopAnimation();

  return 0;
}
